# -*- coding: utf-8 -*-

#  Copyright © 2015-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


from libc.stdint cimport int64_t
from libcpp cimport bool as cbool


ctypedef void CallbackVoid() nogil
ctypedef void CallbackInt(int) nogil
ctypedef void CallbackString(const QString &) nogil
ctypedef void CallbackDebugMessage(const QOpenGLDebugMessage &) nogil


#### QtCore ####

#global
cdef extern from "<QtCore/QtGlobal>":
    ctypedef unsigned char uchar
    ctypedef unsigned int uint
    ctypedef double qreal
    ctypedef int64_t qint64
    
#Qt
cdef extern from "<QtCore/Qt>" namespace "Qt":
    cdef enum FocusReason:
        OtherFocusReason
    cdef enum FocusPolicy:
        StrongFocus
    cdef enum ConnectionType:
        DirectConnection, QueuedConnection
    cdef enum KeyboardModifier:
        ShiftModifier, ControlModifier, KeypadModifier, AltModifier
    ctypedef int KeyboardModifiers
    cdef enum Key:
        Key_Period, Key_Escape, Key_Right, Key_Left, Key_Up, Key_Down,
        Key_Shift, Key_Control, Key_Meta, Key_Alt, Key_AltGr,
        Key_CapsLock, Key_NumLock, Key_ScrollLock
    cdef enum MouseButton:
        LeftButton, RightButton
    cdef enum CursorShape:
        CrossCursor
    cdef enum TransformationMode:
        FastTransformation, SmoothTransformation
    cdef enum AspectRatioMode:
        IgnoreAspectRatio
    cdef enum ImageConversionFlag:
        AutoColor
    cdef enum ApplicationAttribute:
        AA_X11InitThreads
        AA_UseDesktopOpenGL
        AA_UseSoftwareOpenGL
    cdef enum ItemDataRole:
        DisplayRole, UserRole
        
#QByteArray
cdef extern from "<QtCore/QByteArray>":
    cdef cppclass QByteArray:
        QByteArray() nogil
        QByteArray(const char *data) nogil
        const char *data() nogil const
        int size() nogil const
        
#QElapsedTimer
cdef extern from "<QtCore/QElapsedTimer>":
    cdef cppclass QElapsedTimer:
        QElapsedTimer() nogil
        void start() nogil
        cbool hasExpired(qint64 timeout) nogil const
        qint64 restart() nogil
        
#QEvent
cdef extern from "<QtCore/QEvent>":
    cdef cppclass QEvent:
        pass
        
#QFileSystemWatcher
cdef extern from "<QtCore/QFileSystemWatcher>":
    cdef cppclass QFileSystemWatcher:
        QFileSystemWatcher(QObject *parent)
        cbool addPath(const QString &file)
cdef extern from "<QtCore/QFileSystemWatcher>" namespace "QFileSystemWatcher":
    void directoryChanged(const QString &path) nogil
    void fileChanged(const QString &path) nogil
    
#QLibraryInfo
cdef extern from "<QtCore/QLibraryInfo>" namespace "QLibraryInfo":
    QString location(LibraryLocation loc)
    cdef enum LibraryLocation:
        TranslationsPath
        
#QList[QObject*]
cdef extern from "<QtCore/QList>":
    cdef cppclass QObjectList "QList<QObject*> ":
        #int size() nogil const
        #const QObject *at(int i) nogil const
        void reserve(int size) nogil
        void append(const QObject *t) nogil
        
#QList[QOpenGLDebugMessage]
cdef extern from "<QtCore/QList>":
    cdef cppclass OpenGLDebugMessageList "QList<QOpenGLDebugMessage>":
        int size() nogil const
        const QOpenGLDebugMessage& at(int i) nogil const
        
#QStringList
cdef extern from "<QtCore/QStringList>":
    cdef cppclass QStringList:
        int size() const
        const QString &operator[](int i) const
    
#QList[QVariant] == QVariantList
cdef extern from "<QtCore/QList>":
    cdef cppclass QVariantList "QList<QVariant> ":
        QVariantList()
        int size() const
        const QVariant& at(int i) const
        void reserve(int size)
        void append(const QVariant &t)
        
#QList[QUrl] == QUrlList
cdef extern from "<QtCore/QList>":
    cdef cppclass QUrlList "QList<QUrl> ":
        pass
        #QVariantList()
        #int size() const
        const QUrl& at(int i) const
        #void reserve(int size)
        #void append(const QUrl &t)
        
#QLocale
cdef extern from "<QtCore/QLocale>":
    cdef cppclass QLocale:
        QString name() const
cdef extern from "<QtCore/QLocale>" namespace "QLocale":
    QLocale QLocale_system "QLocale::system" ()
        
#QMap[QString,QVariant] == QVariantMap
cdef extern from "<QtCore/QMap>":
    cdef cppclass QVariantMap "QMap<QString, QVariant> ":
        QVariantMap()
        QStringList keys() const
        QVariantList values() const
        
#QMetaMethod
cdef extern from "<QtCore/QMetaMethod>":
    cdef cppclass QMetaMethod:
        QByteArray methodSignature() nogil const
        const char *typeName() nogil const
#QMetaObject
cdef extern from "<QtCore/QMetaObject>":
    cdef cppclass QMetaObject:
        const char *className() nogil const
        int propertyOffset() nogil const
        int propertyCount() nogil const
        QMetaProperty property(int index) nogil const
        int methodOffset() nogil const
        int methodCount() nogil const
        QMetaMethod method(int index) nogil const
#QMetaObject::Connection
cdef extern from "<QtCore/QMetaObject>" namespace "QMetaObject":
    cdef cppclass Connection:
        pass
    cbool invokeMethod(QObject *obj, const char *member, ConnectionType type) nogil
#QMetaProperty
cdef extern from "<QtCore/QMetaProperty>":
    cdef cppclass QMetaProperty:
        const char *name() nogil const
        const char *typeName() nogil const
        
#QMimeData
cdef extern from "<QtCore/QMimeData>":
    cdef cppclass QMimeData:
        QStringList formats() const
        cbool hasColor() const
        QVariant colorData() const
        cbool hasUrls() const
        QUrlList urls() const
        
#QModelIndex
cdef extern from "<QtCore/QModelIndex>":
    cdef cppclass QModelIndex:
        QVariant data(int role) const
        int row() const
        
#QObject
cdef extern from "<QtCore/QObject>":
    cdef cppclass QObject:
        QString objectName() nogil const
        const QMetaObject *metaObject() nogil const
        QVariant property(const char * name) nogil const
        cbool setProperty(const char *name, const QVariant &value) nogil
        cbool isWindowType() nogil const
cdef extern from "<QtCore/QObject>" namespace "QObject":
    Connection connect(const QTimer *sender, CallbackVoid *signal, CallbackVoid slot) nogil
    Connection connect(const QFileSystemWatcher *sender, CallbackString *signal, CallbackString slot) nogil
    Connection connect(const QApplication *sender, CallbackVoid *signal,
                       const QApplication *receiver, CallbackVoid *slot, ConnectionType type) nogil
    
#QPoint
cdef extern from "<QtCore/QPoint>":
    cdef cppclass QPoint:
        int x() nogil const
        int y() nogil const
        
#QPointF
cdef extern from "<QtCore/QPointF>":
    cdef cppclass QPointF:
        QPointF() nogil const
        qreal x() nogil const
        qreal y() nogil const
        void setX(qreal x) nogil
        void setY(qreal y) nogil
        
#QSize
cdef extern from "<QtCore/QSize>":
    cdef cppclass QSize:
        int	height() const
        int	width() const
        
#QStandardPaths
cdef extern from "<QtCore/QStandardPaths>" namespace "QStandardPaths":
    cdef enum StandardLocation:
        PicturesLocation, HomeLocation
    QStringList standardLocations(StandardLocation type)
    
#QString
cdef extern from "<QtCore/QString>":
    cdef cppclass QString:
        QString() nogil
        cbool isEmpty() nogil const
        QByteArray toUtf8() nogil const
cdef extern from "<QtCore/QString>" namespace "QString":
    QString fromUtf8(const char *str, int size) nogil
    QString QString_number "QString::number"(long n) nogil
    
#QTimer
cdef extern from "<QtCore/QTimer>":
    cdef cppclass QTimer:
        QTimer() nogil
        cbool isActive() nogil const
        void setInterval(int msec) nogil
        int interval() nogil const
        void setSingleShot(cbool singleShot) nogil
        void start(int msec) nogil
        void start() nogil
        void stop() nogil
        void deleteLater() nogil
cdef extern from "<QtCore/QTimer>" namespace "QTimer":
    void singleShot(int msec, CallbackVoid func) nogil
    void timeout() nogil
    
#QTranslator
cdef extern from "<QtCore/QTranslator>":
    cdef cppclass QTranslator:
        QTranslator()
        cbool load(const QString &filename, const QString &directory)
        
#QUrl
cdef extern from "<QtCore/QUrl>":
    cdef cppclass QUrl:
        QUrl()
        cbool isLocalFile() const
        QString toLocalFile() const
        QString toString() const
cdef extern from "<QtCore/QUrl>" namespace "QUrl":
    QUrl fromLocalFile(const QString &localfile)
    
#QVariant
cdef extern from "<QtCore/QVariant>":
    cdef cppclass QVariant:
        QVariant() nogil
        QVariant(int val) nogil
        cbool isValid() const
        cbool isNull() const
        void setValue[T](const T &value) nogil
        T value[T]() nogil const
        const char *typeName() nogil const
        int userType() const
        cbool canConvert[T]() nogil const
        QString toString() const
        
        
#### QtGui ####

#QCloseEvent
cdef extern from "<QtGui/QCloseEvent>":
    cdef cppclass QCloseEvent:
        pass
        
#QColor
cdef extern from "<QtGui/QColor>":
    cdef cppclass QColor:
        void setNamedColor(const QString& name)
        int red() const
        int green() const
        int blue() const
        qreal redF() const
        qreal greenF() const
        qreal blueF() const
        
#QCursor
cdef extern from "<QtGui/QCursor>":
    cdef cppclass QCursor:
        QCursor()
        QCursor(const QPixmap &pixmap, int hotX, int hotY)
        void setShape(CursorShape shape)
        
#QDragEnterEvent
cdef extern from "<QtGui/QDragEnterEvent>":
    cdef cppclass QDragEnterEvent:
        void acceptProposedAction()
        const QMimeData *mimeData() const
        
#QDropEvent
cdef extern from "<QtGui/QDropEvent>":
    cdef cppclass QDropEvent:
        void acceptProposedAction()
        const QMimeData *mimeData() const
        QPoint pos() const
        
#QFont
cdef extern from "<QtGui/QFont>":
    cdef cppclass QFont:
        QFont()
        void setFamily(const QString &family)
        void setPixelSize(int pixelSize)
        void setWeight(int weight)
        
#QIcon
cdef extern from "<QtGui/QIcon>":
    cdef cppclass QIcon:
        QIcon()
        QIcon(const QPixmap &pixmap)
        QIcon(const QString &fileName)
        void addFile(const QString &fileName)
        
#QImage
cdef extern from "<QtGui/QImage>" namespace "QImage":
    cdef enum Format:
        Format_RGBA8888
cdef extern from "<QtGui/QImage>":
    cdef cppclass QImage:
        QImage()
        QImage(const uchar *data, int width, int height, Format format)
        QImage(const QString &fileName)
        cbool load(const QString &fileName, const char* format)
        cbool save(const QString &fileName, const char *format, int quality) const
        cbool isNull() const
        int width() const
        int height() const
        QImage scaled(int w, int h, AspectRatioMode aspectMode, TransformationMode mode) const
        QImage mirrored(cbool horizontal, cbool vertical) const
        QImage transformed(const QTransform &matrix) const
        QImage convertToFormat(Format f, ImageConversionFlag flags) const
        const uchar *bits() const
        int byteCount() const
        
#QKeyEvent
cdef extern from "<QtGui/QKeyEvent>":
    cdef cppclass QKeyEvent:
        void ignore()
        KeyboardModifiers modifiers() const
        int key() const
        cbool matches(StandardKey key) const
        
#QKeySequence
cdef extern from "<QtGui/QKeySequence>":
    cdef cppclass QKeySequence:
        QKeySequence()
        QKeySequence(int k1, int k2, int k3, int k4)
        QKeySequence(const QString &key)
        int count() const
        int operator[](uint i) const
        QString toString(SequenceFormat format) const
cdef extern from "<QtGui/QKeySequence>" namespace "QKeySequence":
    cdef enum SequenceFormat:
        PortableText
    cdef enum StandardKey:
        MoveToNextWord, MoveToPreviousWord
    QKeySequence fromString(const QString &str, SequenceFormat format)
        
#QMouseEvent
cdef extern from "<QtGui/QMouseEvent>":
    cdef cppclass QMouseEvent:
        KeyboardModifiers modifiers() const
        MouseButton button() const
        int x() const
        int y() const
        
#QOpenGLContext
cdef extern from "<QtGui/QOpenGLContext>":
    cdef cppclass QOpenGLContext:
        QSurfaceFormat format() nogil const
        cbool isOpenGLES() nogil const
        cbool hasExtension(const QByteArray &extension) nogil const
        QOpenGLFunctions *functions() nogil const
cdef extern from "<QtGui/QOpenGLContext>" namespace "QOpenGLContext":
    QOpenGLContext *currentContext() nogil
    cdef enum OpenGLModuleType:
        LibGL, LibGLES
    OpenGLModuleType openGLModuleType() nogil
    
#QOpenGLDebugLogger
cdef extern from "<QtGui/QOpenGLDebugLogger>":
    cdef cppclass QOpenGLDebugLogger:
        QOpenGLDebugLogger(QObject *parent) nogil
        cbool initialize() nogil
        void enableMessages() nogil
        void startLogging(LoggingMode loggingMode) nogil
        OpenGLDebugMessageList loggedMessages() nogil const
cdef extern from "<QtGui/QOpenGLDebugLogger>" namespace "QOpenGLDebugLogger":
    void messageLogged(const QOpenGLDebugMessage &debugMessage) nogil
    cdef enum LoggingMode:
        AsynchronousLogging, SynchronousLogging
        
#QOpenGLDebugMessage
cdef extern from "<QtGui/QOpenGLDebugMessage>":
    ctypedef unsigned int GLuint
    cdef cppclass QOpenGLDebugMessage:
        GLuint id() nogil const
        QString message() nogil const
        #Severity severity() nogil const
        #Source source() nogil const
        
#QOpenGLFramebufferObjectFormat
cdef extern from "<QtGui/QOpenGLFramebufferObjectFormat>":
    cdef cppclass QOpenGLFramebufferObjectFormat:
        QOpenGLFramebufferObjectFormat() nogil
        int samples() nogil const
        void setSamples(int samples) nogil
        
#QOpenGLFramebufferObject
cdef extern from "<QtGui/QOpenGLFramebufferObject>":
    ctypedef unsigned int GLenum
    cdef cppclass QOpenGLFramebufferObject:
        QOpenGLFramebufferObject(int width, int height, GLenum target) nogil
        QOpenGLFramebufferObject(int width, int height, const QOpenGLFramebufferObjectFormat & format) nogil
        QOpenGLFramebufferObjectFormat format() nogil const
        void setAttachment(Attachment attachment) nogil
        cbool bind() nogil
        cbool release() nogil
        QImage toImage() nogil const
    cdef enum:
        GL_TEXTURE_2D
cdef extern from "<QtGui/QOpenGLFramebufferObject>" namespace "QOpenGLFramebufferObject":
    cdef enum Attachment:
        Depth
        
#QOpenGLFunctions
cdef extern from "<QtGui/QOpenGLFunctions>":
    cdef cppclass QOpenGLFunctions:
        OpenGLFeature openGLFeatures() nogil const
cdef extern from "<QtGui/QOpenGLFunctions>" namespace "QOpenGLFunctions":
    cdef enum OpenGLFeature:
        pass
        
#QOpenGLTexture
cdef extern from "<QtGui/QOpenGLTexture>":
    cdef cppclass QOpenGLTexture:
        QOpenGLTexture(Target target) nogil
        void setFormat(TextureFormat format) nogil
        void setSize(int width, int height, int depth) nogil
        void setMinMagFilters(Filter minificationFilter, Filter magnificationFilter) nogil
        void allocateStorage() nogil
        void bind() nogil
        void release() nogil
        void destroy() nogil
cdef extern from "<QtGui/QOpenGLTexture>" namespace "QOpenGLTexture":
    cdef enum Target:
        Target2D
    cdef enum TextureFormat:
        RGBA32F
    cdef enum Filter:
        Linear
        
#QPaintEvent
cdef extern from "<QtGui/QPaintEvent>":
    cdef cppclass QPaintEvent:
        pass
        
#QPixmap
cdef extern from "<QtGui/QPixmap>":
    cdef cppclass QPixmap:
        QPixmap()
        QPixmap(int width, int height)
        QPixmap(const QString & fileName)
        cbool convertFromImage(const QImage &image)
        cbool save(const QString &fileName, const char *format, int quality) const
cdef extern from "<QtGui/QPixmap>" namespace "QPixmap":
    QPixmap fromImage(const QImage &image)
    
#QResizeEvent
cdef extern from "<QtGui/QResizeEvent>":
    cdef cppclass QResizeEvent:
        const QSize &size() const
        
#QShowEvent
cdef extern from "<QtGui/QShowEvent>":
    cdef cppclass QShowEvent:
        cbool spontaneous() nogil const
        
#QStandardItem
cdef extern from "<QtGui/QStandardItem>":
    cdef cppclass QStandardItem:
        QStandardItem(const QString &text)
        void appendRow(QStandardItem *item)
        QVariant data(int role) const
        void setData(const QVariant &value)
        int row() const
        
#QStandardItemModel
cdef extern from "<QtGui/QStandardItemModel>":
    cdef cppclass QStandardItemModel:
        QStandardItemModel()
        QStandardItemModel(QObject *parent)
        void appendRow(QStandardItem *item)
        void clear()
        int rowCount(const QModelIndex &parent) const
        QModelIndex index(int row, int column, const QModelIndex &parent) const
        QStandardItem *item(int row, int column) const
        void setItem(int row, int column, QStandardItem *item)
        void setHorizontalHeaderItem(int column, QStandardItem *item)
        int rowCount() const
        
#QSurfaceFormat
cdef extern from "<QtGui/QSurfaceFormat>":
    cdef cppclass QSurfaceFormat:
        QSurfaceFormat() nogil
        int depthBufferSize() nogil const
        int stencilBufferSize() nogil const
        int redBufferSize() nogil const
        int greenBufferSize() nogil const
        int blueBufferSize() nogil const
        int alphaBufferSize() nogil const
        void setSamples(int numSamples) nogil
        int samples() nogil const
        SwapBehavior swapBehavior() nogil const
        OpenGLContextProfile profile() nogil const
        void setRenderableType(RenderableType type) nogil
        RenderableType renderableType() nogil const
        int majorVersion() nogil const
        int minorVersion() nogil const
        cbool stereo() nogil const
        void setSwapInterval(int interval) nogil
        int swapInterval() nogil const
        void setOption(FormatOption option) nogil
        FormatOption options() nogil const
cdef extern from "<QtGui/QSurfaceFormat>" namespace "QSurfaceFormat":
    cdef enum SwapBehavior:
        DefaultSwapBehavior, SingleBuffer, DoubleBuffer, TripleBuffer
    cdef enum RenderableType:
        DefaultRenderableType, OpenGL, OpenGLES, OpenVG
    cdef enum OpenGLContextProfile:
        NoProfile, CoreProfile, CompatibilityProfile
    cdef enum FormatOption:
        StereoBuffers, DebugContext
    void setDefaultFormat(const QSurfaceFormat &format) nogil
    QSurfaceFormat defaultFormat() nogil
    
#QTextDocumentFragment
cdef extern from "<QtGui/QTextDocumentFragment>":
    cdef cppclass QTextDocumentFragment:
        QString toHtml() const
cdef extern from "<QtGui/QTextDocumentFragment>" namespace "QTextDocumentFragment":
    QTextDocumentFragment fromPlainText(const QString &plainText)
        
#QTransform
cdef extern from "<QtGui/QTransform>":
    cdef cppclass QTransform:
        QTransform()
        QTransform &rotate(qreal angle)
        
#QWheelEvent
cdef extern from "<QtGui/QWheelEvent>":
    cdef cppclass QWheelEvent:
        QPoint angleDelta() const
        
        
#### QtWidgets ####

#QCoreApplication
cdef extern from "<QtCore/QCoreApplication>":
    cdef cppclass QCoreApplication:
        QStringList arguments()
        void setOrganizationName(const QString &orgName)
        void setApplicationName(const QString &application)
        void setApplicationVersion(const QString &version)
        cbool installTranslator(QTranslator *translationFile)
        int exec()
cdef extern from "<QtCore/QCoreApplication>" namespace "QCoreApplication":
    QCoreApplication *QCoreApplication_instance "QCoreApplication::instance"()
    
#QApplication
cdef extern from "<QtWidgets/QApplication>":
    cdef cppclass QApplication (QCoreApplication):
        QApplication(int &argc, char **argv)
        #QGuiApplication
        void setWindowIcon(const QIcon &icon)
        QFont font()
        void setFont(const QFont &font)
        qreal devicePixelRatio() const
cdef extern from "<QtWidgets/QApplication>" namespace "QApplication":
    void aboutToQuit() nogil
    void setAttribute(ApplicationAttribute attribute) nogil
    void processEvents() nogil
    #QGuiApplication
    void lastWindowClosed() nogil
    void sync() nogil
    
#QWidget
cdef extern from "<QtWidgets/QWidget>":
    cdef cppclass QWidget:
        void deleteLater() nogil
        QObject *sender() const
        QObject* findChild'findChild<QObject*>'(const QString &name) const
        void QWidget_showEvent"QWidget::showEvent"(QShowEvent *event) nogil
        void QWidget_closeEvent"QWidget::closeEvent"(QCloseEvent *event)
        void QWidget_keyPressEvent"QWidget::keyPressEvent"(QKeyEvent *event)
        void QWidget_resizeEvent"QWidget::resizeEvent"(QResizeEvent *event)
        cbool close()
        void setMinimumSize(int minw, int minh) nogil
        void setVisible(cbool visible)
        void resize(int w, int h)
        void show() nogil
        void update()
        void setTitle"setWindowTitle"(const QString &)
        void setWindowTitle(const QString &)
        void setWindowIcon(const QIcon &icon)
        void setFocus()
        QPixmap grab()
        
#QDialog
cdef extern from "<QtWidgets/QDialog>":
    cdef cppclass QDialog(QWidget):
        int exec()
        
#QFileDialog
cdef extern from "<QtWidgets/QFileDialog>" namespace "QFileDialog":
    QString getOpenFileName(QWidget *parent, const QString &caption, const QString &dir)
    
#QLabel
cdef extern from "<QtWidgets/QLabel>":
    cdef cppclass QLabel(QWidget):
        void setText(const QString &)
        
#QLineEdit
cdef extern from "<QtWidgets/QLineEdit>":
    cdef cppclass QLineEdit(QWidget):
        void QLineEdit_keyPressEvent'QLineEdit::keyPressEvent'(QKeyEvent *event)
        QString text() const
        void setText(const QString &)
        int cursorPosition() const
        void setCursorPosition(int)

#QListWidget
cdef extern from "<QtWidgets/QListWidget>":
    cdef cppclass QListWidget(QWidget):
        int	currentRow() const
        
#QListWidgetItem
cdef extern from "<QtWidgets/QListWidgetItem>":
    cdef cppclass QListWidgetItem:
        QListWidget *listWidget() const
        
#QMainWindow
cdef extern from "<QtWidgets/QMainWindow>":
    cdef cppclass QMainWindow(QWidget):
        QStatusBar *statusBar() const
        
#QMessageBox
cdef extern from "<QtWidgets/QMessageBox>":
    cdef cppclass QMessageBox(QWidget):
        QMessageBox(QWidget *parent) nogil
        void setText(const QString &text)
        void setIcon(int)
        void setStandardButtons(int buttons)
        int exec()
cdef extern from "<QtWidgets/QMessageBox>" namespace "QMessageBox":
    cdef enum:
        QMessageBox_Information 'QMessageBox::Information',
        QMessageBox_Close 'QMessageBox::Close',
        
#QOpenGLWidget
cdef extern from "<QtWidgets/QOpenGLWidget>":
    cdef cppclass QOpenGLWidget(QWidget):
        QOpenGLWidget(QWidget *parent) nogil
        int	height() nogil const
        void setAcceptDrops(cbool on) nogil
        void setFocusPolicy(FocusPolicy policy) nogil
        void setMouseTracking(cbool enable) nogil
        #QWidget *window() nogil const
        QSurfaceFormat format() nogil const
        void setSizePolicy(QSizePolicy) nogil
        void paintGL() nogil
        void setCursor(const QCursor &)
        void unsetCursor()
        void makeCurrent()
        void doneCurrent()
        QImage grabFramebuffer()
        
#QPushButton
cdef extern from "<QtWidgets/QPushButton>":
    cdef cppclass QPushButton:
        pass
        
#QSizePolicy
cdef extern from "<QtWidgets/QSizePolicy>":
    cdef cppclass QSizePolicy:
        QSizePolicy()
        QSizePolicy(Policy horizontal, Policy vertical)
        void setHorizontalStretch(int stretchFactor)
        void setVerticalStretch(int stretchFactor)
cdef extern from "<QtWidgets/QSizePolicy>" namespace "QSizePolicy":
    cdef enum Policy:
        Expanding
        
#QStatusBar
cdef extern from "<QtWidgets/QStatusBar>":
    cdef cppclass QStatusBar:
        void showMessage(const QString &message)
        void addWidget(QWidget *widget, int stretch)
        
#QTreeView
cdef extern from "<QtWidgets/QTreeView>":
    cdef cppclass QTreeView:
        pass
        
    
    




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2014-2015, 2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys
from math import sqrt, atan2


epsdigits = 5
epsilon = 10**-epsdigits
filebyteorder = 'little'

def get_texcoords_range(verts, normal):
    texcoords = [vert.rotationyx_normal(normal) for vert in verts]
    minx = min(x for x, y in texcoords)
    maxx = max(x for x, y in texcoords)
    miny = min(y for x, y in texcoords)
    maxy = max(y for x, y in texcoords)
    return minx, maxx, miny, maxy
    
    
class Coords (tuple):
    def __new__(cls, args=(0,0,0)):
        return tuple.__new__(cls, args)
        
    def __add__(self, other):
        return self.__class__(s+o for s, o in zip(self, other))
        
    def __sub__(self, other):
        return self.__class__(s-o for s, o in zip(self, other))
        
    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, super().__repr__())
    def __str__(self):
        return str(list(self))
        
    
class Vector(Coords):
    def __neg__(self):
        return self.__class__(-s for s in self)
        
    def __mul__(self, value):
        return self.__class__(s*value for s in self)
        
    def __truediv__(self, value):
        return self.__class__(s/value for s in self)
        
    def cross(self, other):
        return self.__class__((
                    self[1] * other[2] - other[1] * self[2],
                    self[2] * other[0] - other[2] * self[0],
                    self[0] * other[1] - other[0] * self[1]))
                    
    def dot(self, other): return sum(s*o for s, o in zip(self, other))
    def length_squared(self): return self.dot(self)
    def length(self): return sqrt(self.length_squared())
    def normalised(self): return self / self.length()
    
    def angle(self, other):
        sina = self.cross(other).length()
        cosa = self.dot(other)
        return atan2(sina, cosa)
        
    def angle_plane(self, other, plane_normal):
        cross = self.cross(other)
        sina = cross.length()
        cosa = self.dot(other)
        angle = atan2(sina, cosa)
        if cross.dot(plane_normal) < 0:
            return - angle
        return angle
        
    def anglex(self):
        sina = -self[1]
        cosa = -self[2]
        if sina*sina + cosa*cosa < epsilon:
            cosa, sina = 1, 0
        return cosa, sina
        
    def angley(self):
        sina = -self[0]
        cosa = self[2]
        if sina*sina + cosa*cosa < epsilon:
            cosa, sina = 1, 0
        return cosa, sina
        
    def rotationx(self, cosa, sina):
        return self.__class__((self[0], cosa*self[1] - sina*self[2], sina*self[1] + cosa*self[2]))
        
    def rotationy(self, cosa, sina):
        return self.__class__((cosa*self[0] + sina*self[2], self[1], cosa*self[2] - sina*self[0]))
        
    def angleyx(self):
        ''' Same as:
            cosay, sinay = self.angley()
            cosax, sinax = self.rotationy(cosay, sinay).anglex()
            return cosay, sinay, cosax, sinax
        '''
        x, y, z = self
        xx = x*x
        zz = z*z
        xx_zz = xx + zz
        if xx_zz < epsilon:
            if y*y + zz < epsilon:
                return 1, 0, 1, 0
            else:
                return 1, 0, -z, -y
        else:
            if y*y + xx_zz*xx_zz < epsilon:
                return z, -x, 1, 0
            else:
                return z, -x, -xx_zz, -y
                
    def rotationyx(self, cosay, sinay, cosax, sinax):
        x, y, z = self
        x, z = cosay*x + sinay*z, cosay*z - sinay*x
        y, z = cosax*y - sinax*z, sinax*y + cosax*z
        return self.__class__((x, y, z))
        
    def rotationyx_normal(self, normal):
        ''' Same as:
            self.rotationyx(*normal.angleyx())
        '''
        vx, vy, vz = self
        nx, ny, nz = normal
        nxx = nx*nx
        nzz = nz*nz
        nxx_zz = nxx + nzz
        if nxx_zz < epsilon:
            if ny*ny + nzz < epsilon:
                pass
            else:
                vy = ny*vz - nz*vy
        else:
            if ny*ny + nxx_zz*nxx_zz < epsilon:
                vx = nz*vx - nx*vz
            else:
                vy = ny*(nz*vz + nx*vx) - nxx_zz*vy
                vx = nz*vx - nx*vz
        return vx, vy
        
        
def exc_end(func, *, name=None):
    def _func(*args):
        try:
            return func(*args)
        except:
            sys.excepthook(*sys.exc_info())
    _func.__doc__ = func.__doc__
    _func.__module__ = func.__module__
    _func.__name__ = name or func.__name__
    return _func
    
class ExceptionMeta (type):
    def __new__(cls, name, bases, namespace):
        for k, v in sorted(namespace.items()):
            if k.startswith('on_'):
                namespace[k] = exc_end(v, name=k)
        return type(name, bases, namespace)
        




#-*- coding:utf-8 -*-

#  Copyright © 2009, 2011-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys
import importlib
from collections import namedtuple, OrderedDict

from .debug import debug, DEBUG_MSG, DEBUG_ALG, DEBUG_LIVEPLUGINS
from . import config
from . import model

FILE_FORMAT = 'Pybik Plugins Collection - Version 2.0'
FILE_SIG = 'Format: ' + FILE_FORMAT


class PluginFileError (Exception): pass
class PluginSolverAbort (Exception): pass
class PluginModelCompatError (Exception): pass


class PluginFileLoader:
    def __init__(self, filename):
        #                       field name:     (required, singleline, convert-func)
        parse_pass = lambda value, lineno, comment: value
        self.header_scheme = {  'Format':       (True,     True,       self.parse_version),
                                'Copyright':    (False,    False,      parse_pass),
                                'License':      (False,    False,      parse_pass),
                                'Model':        (False,    False,      self.parse_model),
                                'Ref-Blocks':   (False,    True,       self.parse_refblocks),
                            }
        self.body_scheme =  {   'Path':         (True,     True,       self.parse_path),
                                'Depends':      (False,    False,      parse_pass),
                                'Model':        (True,     False,      self.parse_model),
                                'Ref-Blocks':   (False,    True,       self.parse_refblocks),
                                'Solution':     (False,    False,      self.parse_solution),
                                'Moves':        (False,    False,      self.parse_moves),
                                'Module':       (False,    True,       self.parse_module),
                            }
        self.conflicts = [('Solution', 'Moves', 'Module'), ('Depends', 'Moves', 'Module')]
                            
        self.paras = []
        self.header = None
        self.body = []
        self.load(filename)
        self.parse()
        
    def load(self, filename):
        debug('loading plugin-file', repr(filename))
        with open(filename, 'rt', encoding='utf-8') as fp:
            lines = fp.readlines()
        
        # parse file
        para = {}
        key = None
        comment = None
        for lineno, line in enumerate(lines):
            line = line.rstrip()
            if not line:
                # end para
                if not self.paras and not para:
                    raise PluginFileError('File must not start with empty line.')
                self.paras.append(para)
                para = {}
                comment = None
            elif line[0] == '#':
                # comment
                if not self.paras and not para:
                    raise PluginFileError('File must not start with comment.')
                comment = line
            elif line[0] in ' \t':
                if not para:
                    raise PluginFileError('Paragraph must not start with indented line.')
                line = line.strip()
                if line[0] != '#':
                    # multiline
                    para[key][0].append(line)
                comment = None
            else:
                if not self.paras and not para and line.rstrip() != FILE_SIG:
                    raise PluginFileError('File must start with:', FILE_SIG)
                key, value = line.split(':')
                value = value.strip()
                para[key] = [value] if value else [], lineno, comment
                comment = None
        if para:
            self.paras.append(para)
        
    @staticmethod
    def parse_version(value, unused_lineno, unused_comment):
        if value != FILE_FORMAT:
            raise PluginFileError('wrong file version:', value)
        return value
        
    @staticmethod
    def parse_model(modelstrings, unused_lineno, unused_comment):
        model_infos = []
        for modelstr in modelstrings or []:
            try:
                model_info = model.Model.from_string(modelstr)
            except ValueError as e:
                debug('Error in model %s:' % modelstr, e)
            else:
                model_infos.append(model_info)
        return model_infos
        
    @staticmethod
    def parse_refblocks(value, unused_lineno, unused_comment):
        return value and [[v, v] for v in value.upper().split()]
        
    @staticmethod
    def parse_path(value, lineno, comment):
        if not value:
            raise PluginFileError('empty path')
        sep = value[0]
        if sep == '@':
            return value, None, lineno, comment
        def _translate(value):
            transl = True
            for v in value.strip(sep).split(sep):
                if not v:
                    continue
                if transl is True:
                    if v == '_':    transl = True
                    elif v == 'P_': transl = 'sing'
                    elif v == '!_': transl = False
                    else:           yield v, _(v)
                elif transl is False:
                    yield v, v
                    transl = True
                elif transl == 'sing':
                    vs = v
                    transl = 'plural'
                elif transl == 'plural':
                    vp = v
                    transl = 'n'
                elif transl == 'n':
                    try:
                        vi = int(v)
                    except ValueError:
                        raise PluginFileError('Third value after P_ must be int.')
                    yield (vs if vi==1 else vp).format(v), ngettext(vs, vp, vi).format(v)
        return value, tuple(_translate(value)), lineno, comment
        
    @staticmethod
    def parse_solution(value, unused_lineno, unused_comment):
        def split_solution(value):
            try:
                conditions, moves = value.split(',')
            except ValueError:
                raise PluginFileError('expected exactly one comma in: ' + value)
            conditions = conditions.strip().upper().split()
            conditions = [c.split('=') for c in conditions]
            for c in conditions:
                if len(c) != 2:
                    raise PluginFileError('expected exactly one "=" in: ' + '='.join(c))
            return value, conditions, moves.strip()
        return value and [split_solution(v) for v in value if v]
        
    @staticmethod
    def parse_moves(value, unused_lineno, unused_comment):
        if value is None:
            return None
        moves = ' '.join(value)
        def func(game):
            game.move_sequence.reset()
            game.add_code(moves)
            game.set_plugin_mode('append')
        return func
        
    @staticmethod
    def parse_module(value, unused_lineno, unused_comment):
        if value is None:
            return None
        value = value.split(',')
        if len(value) != 2:
            raise PluginFileError('Wrong Syntax in Module field, expected: module.func, flag')
        value, flag = value
        modulename, funcname = value.rstrip().rsplit('.', 1)
        from . import plugins
        moduleobj = importlib.import_module('.'+modulename, plugins.__package__)
        modulefunc = getattr(moduleobj, funcname)
        flag = flag.strip()
        if flag not in ['append', 'replace', 'challenge']:
            raise PluginFileError('Unknown flag %s for module' % flag)
        def func(game):
            if flag == 'append':
                game.move_sequence.reset()
            modulefunc(game)
            game.set_plugin_mode(flag)
        return func
        
    def parse(self):
        if not self.paras:
            raise PluginFileError('empty plugin file')
        self.header = self.paras.pop(0)
        default = {k:v for k,v in self.header.items() if k in self.header_scheme and k in self.body_scheme}
        self.parse_para(self.header, self.header_scheme)
        
        def check_conflicts(para):
            for conflicts in self.conflicts:
                if sum(int(c in para) for c in conflicts) > 1:
                    debug('Only one field allowed of:', ', '.join(conflicts))
                    debug('Skipping paragraph %s in plugin file' % para.get('Path'))
                    return False
            return True
            
        while self.paras:
            para = self.paras.pop(0)
            if not para:
                continue
            if DEBUG_ALG: debug('    Path:', para.get('Path'))
            if not check_conflicts(para):
                continue
            for k, v in default.items():
                para.setdefault(k, v)
            try:
                self.parse_para(para, self.body_scheme)
            except PluginFileError as e:
                debug('Skipping paragraph in plugin file:', e, 'keys:', *para.keys())
            else:
                self.body.append(para)
        self.paras = None
                
    @staticmethod
    def parse_para(para, scheme):
        for key, (required, singleline, convert_func) in scheme.items():
            try:
                value, lineno, comment = para[key]
            except KeyError:
                if required:
                    raise PluginFileError('missing required field {!r}'.format(key))
                else:
                    value, lineno, comment = None, None, None
            if singleline and value is not None:
                if len(value) == 0:
                    value = None
                elif len(value) == 1:
                    value = value[0]
                else:
                    raise PluginFileError('{!r} is a singleline field, but {} lines found'.format(key, len(value)))
            para[key] = convert_func(value, lineno, comment)
            
            
class PluginHelper:
    def __init__(self):
        self.scripts = OrderedDict()
        
    @staticmethod
    def test_model(model, models):
        for unused_modelstr, mtype, sizes, exp in models:
            if mtype == '*':
                return True
            if mtype is None:
                continue
            if model.type != mtype:
                continue
            sizedict = {}
            for size1, size2 in zip(sizes, model.sizes):
                if size1 is None:
                    continue
                if type(size1) is int:
                    if size1 != size2:
                        break
                else:
                    sizedict[size1] = size2
            else:
                try:
                    if exp is None or eval(exp, {}, sizedict):
                        return True
                except ValueError:
                    continue
        return False
        
    def get_function(self, model, index):
        all_models = []
        for models, func in list(self.scripts.values())[index]:
            if self.test_model(model, models):
                assert func is not None
                return func
            all_models += models
        else:
            # all model rejected
            all_models = [model_[0] for model_ in all_models if model_[1] is not None]
            if not all_models:
                raise PluginModelCompatError(_('This plugin does not work for any model.')+'\n')
            elif len(all_models) == 1:
                raise PluginModelCompatError(_('This plugin only works for:') + '\n  ' + all_models[0])
            else:
                raise PluginModelCompatError(
                        _('This plugin only works for:') + '\n' +
                        '\n'.join([' • ' + m for m in all_models])
                    )
        
    def load_plugins_from_directory(self, dirname):
        if DEBUG_MSG or DEBUG_LIVEPLUGINS:
            print('loading plugins from', repr(dirname))
        for filename in sorted(os.listdir(dirname), key=str.lower):
            unused_name, ext = os.path.splitext(filename)
            if ext != '.plugins':
                continue
            if DEBUG_ALG: debug('  plugins:', filename)
            try:
                self.load_plugins_from_file(os.path.join(dirname, filename))
            except Exception:
                sys.excepthook(*sys.exc_info())
        
    def load_plugins(self):
        self.scripts.clear()
        for dirname in [config.PLUGINS_DIR, config.get_data_home(config.PACKAGE, 'plugins')]:
            if os.path.isdir(dirname):
                debug('Found plugins path:', dirname)
                self.load_plugins_from_directory(dirname)
            else:
                debug('Plugins path does not exist:', dirname)
        if DEBUG_ALG: debug('found', len(self.scripts))
        return [(path, i) for i, path in enumerate(self.scripts.keys())]
        
    @staticmethod
    def resolve_depends(params):
        depends = []
        # resolve dependencies to other scripts
        for depend in params.depends:
            for path, unused_models, instance in params.scripts:
                if path == depend:
                    depends += instance.params.depends
            depends.append(depend)
        if DEBUG_ALG:
            debug('    Resolve dependencies for:', params.path)
            if params.depends != depends:
                debug('      before:', params.depends)
                debug('      after: ', depends)
        params.depends[:] = depends
        # resolve dependencies to params
        for path, unused_models, instance in params.scripts:
            depends = []
            for depend in instance.params.depends:
                if params.path == depend:
                    depends += params.depends
                depends.append(depend)
            if DEBUG_ALG:
                debug('      Resolve dependencies for:', path)
                if instance.params.depends != depends:
                    debug('        before:', instance.params.depends)
                    debug('        after: ', depends)
            instance.params.depends[:] = depends
            
    def load_plugins_from_file(self, filename, solutionfactory=None):
        try:
            file = PluginFileLoader(filename)
        except PluginFileError as e:
            print('Skipping plugin file:', e)
            return
            
        scripts = []
        for para in file.body:
            path, pathitems, *unused = para['Path']
            models = para['Model']
            depends = para['Depends']
            solution = para['Solution']
            moves = para['Moves']
            module = para['Module']
            if depends or solution:
                params = ScriptParams(
                                depends=depends or [],
                                precond=para['Ref-Blocks'] or [],
                                solution=solution,
                                scripts=scripts,
                                path=path,
                            )
                self.resolve_depends(params)
                func = (solutionfactory or Solution)(params)
            elif moves is not None:
                func = moves
            elif module is not None:
                func = module
            else:
                debug('    skip Path without plugin:', path)
                continue
            scripts.append((path, models, func))
            if pathitems is not None:
                funclist = self.scripts.setdefault(pathitems, [])
                funclist.append((models, func))
        
        
ScriptParams = namedtuple('ScriptParams', 'depends precond solution scripts path')

class Solution:
    def __init__(self, params):
        self.solved_face_colors = {}
        self.params = params
        
    def __call__(self, game):
        game.move_sequence.reset()
        scripts = {path: func for path, models, func in self.params.scripts}
        for depend in self.params.depends:
            instance = scripts[depend]
            self.execute(game, instance.params)
        self.execute(game, self.params)
        game.set_plugin_mode('append')
        
    def test_face(self, cube, blockpos, position, condition):
        color1 = cube.get_colorsymbol(blockpos, position)
        color2 = self.solved_face_colors.setdefault(condition, color1)
        return color1 == color2
        
    def test_basic_condition(self, cube, position, condition):
        assert len(position) == len(condition)
        blockpos = cube.model.blocksym_to_blockpos(position)
        for pos, cond in zip(position, condition):
            if not self.test_face(cube, blockpos, pos, cond):
                return False
        return True
        
    @staticmethod
    def opposite(face):
        #FIXME: this only works for BrickModel
        return {  'F': 'B', 'B': 'F',
                  'L': 'R', 'R': 'L',
                  'U': 'D', 'D': 'U',
               }[face]
        
    def test_pattern_condition(self, cube, position, condition):
        if '?' in condition:
            conditions = (condition.replace('?', face, 1)
                            for face in 'FLUBRD'
                                if face not in condition
                                if self.opposite(face) not in condition)
            return self.test_or_conditions(cube, position, conditions)
        else:
            return self.test_basic_condition(cube, position, condition)
            
    @staticmethod
    def rotated_conditions(condition):
        for i in range(len(condition)):
            yield condition[i:] + condition[:i]
        
    def test_prefix_condition(self, cube, position, condition):
        if condition.startswith('!*'):
            return not self.test_or_conditions(cube, position, self.rotated_conditions(condition[2:]))
        elif condition.startswith('*'):
            #TODO: Instead of testing only rotated conditions, test all permutations.
            #      This should not break existing rules, and would allow to match
            #      e.g. dfr and dfl. Could be done by comparing sorted strings after
            #      expanding patterns.
            return self.test_or_conditions(cube, position, self.rotated_conditions(condition[1:]))
        elif condition.startswith('!'):
            return not self.test_pattern_condition(cube, position, condition[1:])
        else:
            return self.test_pattern_condition(cube, position, condition)
        
    def test_or_conditions(self, cube, position, conditions):
        for prefix_cond in conditions:
            if self.test_prefix_condition(cube, position, prefix_cond):
                return True
        return False
        
    def test_and_conditions(self, cube, conditions):
        for position, or_cond in conditions:
            if not self.test_or_conditions(cube, position, or_cond.split('|')):
                return False
        return True
        
    def execute(self, game, params):
        rules = params.solution
        if rules is None:
            return
        cube = game.current_state
        count = 0
        pos = 0
        if DEBUG_ALG:
            print('{}:'.format(params.path))
            print(' ', cube)
        while pos < len(rules):
            self.solved_face_colors.clear()
            rule, conditions, moves = rules[pos]
            if self.test_and_conditions(cube, params.precond + conditions):
                if DEBUG_ALG:
                    print('  accept: {:2}. {}'.format(pos+1, rule))
                if moves == '@@solved':
                    return
                if moves == '@@unsolvable':
                    raise PluginSolverAbort(_('This puzzle is not solvable.'))
                if count > 4 * len(rules): # this value is just guessed
                    raise PluginSolverAbort(
                        'An infinite loop was detected. '
                        'This is probably an error in the solution.')
                count += 1
                
                game.add_code(moves)
                pos = 0
            else:
                if DEBUG_ALG:
                    print('  reject: {:2}. {}'.format(pos+1, rule))
                pos += 1
        raise PluginSolverAbort(
            'No matching rules found. '
            'This is probably an error in the solution.')
        
        



# -*- coding: utf-8 -*-
# generated file

languages = (
{'ast': {'ast': 'asturianu',
         'bg': 'búlgaru',
         'bs': 'bosniu',
         'cs': 'checu',
         'de': 'alemán',
         'el': 'griegu',
         'en_GB': 'inglés (Reinu Xuníu)',
         'es': 'español',
         'fi': 'finlandés',
         'fr': 'francés',
         'gl': 'gallegu',
         'he': 'hebréu',
         'it': 'italianu',
         'kab': 'kabileñu',
         'ms': 'malayu',
         'pl': 'polacu',
         'pt_BR': 'portugués (Brasil)',
         'ru': 'rusu',
         'uk': 'ucraín',
         'uz': 'uzbequistanín',
         'zh_TW': 'chinu (Taiwán)'},
 'bg': {'ast': 'астурски',
        'bg': 'български',
        'bs': 'босненски',
        'cs': 'чешки',
        'de': 'немски',
        'el': 'гръцки',
        'en_GB': 'английски (Обединеното кралство)',
        'es': 'испански',
        'fi': 'фински',
        'fr': 'френски',
        'gl': 'галисийски',
        'he': 'иврит',
        'it': 'италиански',
        'kab': 'кабилски',
        'ms': 'малайски',
        'pl': 'полски',
        'pt_BR': 'португалски (Бразилия)',
        'ru': 'руски',
        'uk': 'украински',
        'uz': 'узбекски',
        'zh_TW': 'китайски (Тайван)'},
 'bs': {'ast': 'asturijski',
        'bg': 'bugarski',
        'bs': 'bosanski',
        'cs': 'češki',
        'de': 'njemački',
        'el': 'grčki',
        'en_GB': 'engleski (Velika Britanija)',
        'es': 'španski',
        'fi': 'finski',
        'fr': 'francuski',
        'gl': 'galski',
        'he': 'hebrejski',
        'it': 'italijanski',
        'kab': 'kabile',
        'ms': 'malajski',
        'pl': 'poljski',
        'pt_BR': 'portugalski (Brazil)',
        'ru': 'ruski',
        'uk': 'ukrajinski',
        'uz': 'uzbečki',
        'zh_TW': 'kineski (Tajvan)'},
 'cs': {'ast': 'asturština',
        'bg': 'bulharština',
        'bs': 'bosenština',
        'cs': 'čeština',
        'de': 'němčina',
        'el': 'řečtina',
        'en_GB': 'angličtina (Velká Británie)',
        'es': 'španělština',
        'fi': 'finština',
        'fr': 'francouzština',
        'gl': 'galicijština',
        'he': 'hebrejština',
        'it': 'italština',
        'kab': 'kabylština',
        'ms': 'malajština',
        'pl': 'polština',
        'pt_BR': 'portugalština (Brazílie)',
        'ru': 'ruština',
        'uk': 'ukrajinština',
        'uz': 'uzbečtina',
        'zh_TW': 'čínština (Tchaj-wan)'},
 'de': {'ast': 'Asturianisch',
        'bg': 'Bulgarisch',
        'bs': 'Bosnisch',
        'cs': 'Tschechisch',
        'de': 'Deutsch',
        'el': 'Griechisch',
        'en_GB': 'Englisch (Vereinigtes Königreich)',
        'es': 'Spanisch',
        'fi': 'Finnisch',
        'fr': 'Französisch',
        'gl': 'Galizisch',
        'he': 'Hebräisch',
        'it': 'Italienisch',
        'kab': 'Kabylisch',
        'ms': 'Malaiisch',
        'pl': 'Polnisch',
        'pt_BR': 'Portugiesisch (Brasilien)',
        'ru': 'Russisch',
        'uk': 'Ukrainisch',
        'uz': 'Usbekisch',
        'zh_TW': 'Chinesisch (Taiwan)'},
 'el': {'ast': 'Αστουριανά',
        'bg': 'Βουλγαρικά',
        'bs': 'Βοσνιακά',
        'cs': 'Τσεχικά',
        'de': 'Γερμανικά',
        'el': 'Ελληνικά',
        'en_GB': 'Αγγλικά (Ηνωμένο Βασίλειο)',
        'es': 'Ισπανικά',
        'fi': 'Φινλανδικά',
        'fr': 'Γαλλικά',
        'gl': 'Γαλικιανά',
        'he': 'Εβραϊκά',
        'it': 'Ιταλικά',
        'kab': 'Καμπίλε',
        'ms': 'Μαλάι',
        'pl': 'Πολωνικά',
        'pt_BR': 'Πορτογαλικά (Βραζιλία)',
        'ru': 'Ρωσικά',
        'uk': 'Ουκρανικά',
        'uz': 'Ουζμπεκικά',
        'zh_TW': 'Κινεζικά (Ταϊβάν)'},
 'en_GB': {'ast': 'Asturian',
           'bg': 'Bulgarian',
           'bs': 'Bosnian',
           'cs': 'Czech',
           'de': 'German',
           'el': 'Greek',
           'en_GB': 'English (United Kingdom)',
           'es': 'Spanish',
           'fi': 'Finnish',
           'fr': 'French',
           'gl': 'Galician',
           'he': 'Hebrew',
           'it': 'Italian',
           'kab': 'Kabyle',
           'ms': 'Malay',
           'pl': 'Polish',
           'pt_BR': 'Portuguese (Brazil)',
           'ru': 'Russian',
           'uk': 'Ukrainian',
           'uz': 'Uzbek',
           'zh_TW': 'Chinese (Taiwan)'},
 'es': {'ast': 'asturiano',
        'bg': 'búlgaro',
        'bs': 'bosnio',
        'cs': 'checo',
        'de': 'alemán',
        'el': 'griego',
        'en_GB': 'inglés (Reino Unido)',
        'es': 'español',
        'fi': 'finés',
        'fr': 'francés',
        'gl': 'gallego',
        'he': 'hebreo',
        'it': 'italiano',
        'kab': 'cabila',
        'ms': 'malayo',
        'pl': 'polaco',
        'pt_BR': 'portugués (Brasil)',
        'ru': 'ruso',
        'uk': 'ucraniano',
        'uz': 'uzbeko',
        'zh_TW': 'chino (Taiwán)'},
 'fi': {'ast': 'asturia',
        'bg': 'bulgaria',
        'bs': 'bosnia',
        'cs': 'tšekki',
        'de': 'saksa',
        'el': 'kreikka',
        'en_GB': 'englanti (Iso-Britannia)',
        'es': 'espanja',
        'fi': 'suomi',
        'fr': 'ranska',
        'gl': 'galicia',
        'he': 'heprea',
        'it': 'italia',
        'kab': 'kabyyli',
        'ms': 'malaiji',
        'pl': 'puola',
        'pt_BR': 'portugali (Brasilia)',
        'ru': 'venäjä',
        'uk': 'ukraina',
        'uz': 'uzbekki',
        'zh_TW': 'kiina (Taiwan)'},
 'fr': {'ast': 'asturien',
        'bg': 'bulgare',
        'bs': 'bosniaque',
        'cs': 'tchèque',
        'de': 'allemand',
        'el': 'grec',
        'en_GB': 'anglais (Royaume-Uni)',
        'es': 'espagnol',
        'fi': 'finnois',
        'fr': 'français',
        'gl': 'galicien',
        'he': 'hébreu',
        'it': 'italien',
        'kab': 'kabyle',
        'ms': 'malais',
        'pl': 'polonais',
        'pt_BR': 'portugais (Brésil)',
        'ru': 'russe',
        'uk': 'ukrainien',
        'uz': 'ouzbek',
        'zh_TW': 'chinois (Taïwan)'},
 'gl': {'ast': 'asturiano',
        'bg': 'búlgaro',
        'bs': 'bosnio',
        'cs': 'checo',
        'de': 'alemán',
        'el': 'grego',
        'en_GB': 'inglés (Reino Unido)',
        'es': 'español',
        'fi': 'finés',
        'fr': 'francés',
        'gl': 'galego',
        'he': 'hebreo',
        'it': 'italiano',
        'kab': 'kabile',
        'ms': 'malaio',
        'pl': 'polaco',
        'pt_BR': 'portugués (Brasil)',
        'ru': 'ruso',
        'uk': 'ucraíno',
        'uz': 'uzbeco',
        'zh_TW': 'chinés (Taiwán)'},
 'he': {'ast': 'אסטורית',
        'bg': 'בולגרית',
        'bs': 'בוסנית',
        'cs': 'צ׳כית',
        'de': 'גרמנית',
        'el': 'יוונית',
        'en_GB': 'אנגלית (הממלכה המאוחדת)',
        'es': 'ספרדית',
        'fi': 'פינית',
        'fr': 'צרפתית',
        'gl': 'גליציאנית',
        'he': 'עברית',
        'it': 'איטלקית',
        'kab': 'קבילה',
        'ms': 'מלאית',
        'pl': 'פולנית',
        'pt_BR': 'פורטוגלית (ברזיל)',
        'ru': 'רוסית',
        'uk': 'אוקראינית',
        'uz': 'אוזבקית',
        'zh_TW': 'סינית (טייוואן)'},
 'it': {'ast': 'asturiano',
        'bg': 'bulgaro',
        'bs': 'bosniaco',
        'cs': 'ceco',
        'de': 'tedesco',
        'el': 'greco',
        'en_GB': 'inglese (Regno Unito)',
        'es': 'spagnolo',
        'fi': 'finlandese',
        'fr': 'francese',
        'gl': 'galiziano',
        'he': 'ebraico',
        'it': 'italiano',
        'kab': 'cabilo',
        'ms': 'malese',
        'pl': 'polacco',
        'pt_BR': 'portoghese (Brasile)',
        'ru': 'russo',
        'uk': 'ucraino',
        'uz': 'usbeco',
        'zh_TW': 'cinese (Taiwan)'},
 'kab': {'bg': 'Tabulgarit',
         'cs': 'Tačikit',
         'de': 'Talmant',
         'el': 'Tagrikit',
         'en_GB': 'Taglizit (Tagelda Yedduklen)',
         'es': 'Taspenyulit',
         'fr': 'Tafransist',
         'it': 'Taṭalyanit',
         'kab': 'Taqbaylit',
         'ms': 'Tamalawit',
         'pl': 'Tapulunit',
         'pt_BR': 'Tapurtugalit (Brizil)',
         'ru': 'Tarusit',
         'uk': 'Tukranit',
         'zh_TW': 'Tacinwat, Tamundarint (Ṭaywan)'},
 'ms': {'bg': 'Bulgaria',
        'bs': 'Bosnia',
        'cs': 'Czech',
        'de': 'Jerman',
        'el': 'Greek',
        'en_GB': 'Inggeris (United Kingdom)',
        'es': 'Sepanyol',
        'fi': 'Finland',
        'fr': 'Perancis',
        'gl': 'Galicia',
        'he': 'Ibrani',
        'it': 'Itali',
        'kab': 'Kabyle',
        'ms': 'Bahasa Melayu',
        'pl': 'Poland',
        'pt_BR': 'Portugis (Brazil)',
        'ru': 'Rusia',
        'uk': 'Ukraine',
        'uz': 'Uzbekistan',
        'zh_TW': 'Cina (Taiwan)'},
 'pl': {'ast': 'asturyjski',
        'bg': 'bułgarski',
        'bs': 'bośniacki',
        'cs': 'czeski',
        'de': 'niemiecki',
        'el': 'grecki',
        'en_GB': 'angielski (Wielka Brytania)',
        'es': 'hiszpański',
        'fi': 'fiński',
        'fr': 'francuski',
        'gl': 'galicyjski',
        'he': 'hebrajski',
        'it': 'włoski',
        'kab': 'kabylski',
        'ms': 'malajski',
        'pl': 'polski',
        'pt_BR': 'portugalski (Brazylia)',
        'ru': 'rosyjski',
        'uk': 'ukraiński',
        'uz': 'uzbecki',
        'zh_TW': 'chiński (Tajwan)'},
 'pt_BR': {'ast': 'asturiano',
           'bg': 'búlgaro',
           'bs': 'bósnio',
           'cs': 'tcheco',
           'de': 'alemão',
           'el': 'grego',
           'en_GB': 'inglês (Reino Unido)',
           'es': 'espanhol',
           'fi': 'finlandês',
           'fr': 'francês',
           'gl': 'galego',
           'he': 'hebraico',
           'it': 'italiano',
           'kab': 'kabyle',
           'ms': 'malaio',
           'pl': 'polonês',
           'pt_BR': 'português (Brasil)',
           'ru': 'russo',
           'uk': 'ucraniano',
           'uz': 'usbeque',
           'zh_TW': 'chinês (Taiwan)'},
 'ru': {'ast': 'астурийский',
        'bg': 'болгарский',
        'bs': 'боснийский',
        'cs': 'чешский',
        'de': 'немецкий',
        'el': 'греческий',
        'en_GB': 'английский (Великобритания)',
        'es': 'испанский',
        'fi': 'финский',
        'fr': 'французский',
        'gl': 'галисийский',
        'he': 'иврит',
        'it': 'итальянский',
        'kab': 'кабильский',
        'ms': 'малайский',
        'pl': 'польский',
        'pt_BR': 'португальский (Бразилия)',
        'ru': 'русский',
        'uk': 'украинский',
        'uz': 'узбекский',
        'zh_TW': 'китайский (Тайвань)'},
 'uk': {'ast': 'астурська',
        'bg': 'болгарська',
        'bs': 'боснійська',
        'cs': 'чеська',
        'de': 'німецька',
        'el': 'грецька',
        'en_GB': 'англійська (Велика Британія)',
        'es': 'іспанська',
        'fi': 'фінська',
        'fr': 'французька',
        'gl': 'галісійська',
        'he': 'іврит',
        'it': 'італійська',
        'kab': 'кабільська',
        'ms': 'малайська',
        'pl': 'польська',
        'pt_BR': 'португальська (Бразилія)',
        'ru': 'російська',
        'uk': 'українська',
        'uz': 'узбецька',
        'zh_TW': 'китайська (Тайвань)'},
 'uz': {'bg': 'bolgar',
        'bs': 'bosniy',
        'cs': 'chex',
        'de': 'nemischa',
        'el': 'grek',
        'en_GB': 'inglizcha (Birlashgan Qirollik)',
        'es': 'ispancha',
        'fi': 'fincha',
        'fr': 'fransuzcha',
        'gl': 'galisiy',
        'he': 'ibroniy',
        'it': 'italyan',
        'kab': 'kabil',
        'ms': 'malay',
        'pl': 'polyakcha',
        'pt_BR': 'portugalcha (Braziliya)',
        'ru': 'ruscha',
        'uk': 'ukrain',
        'uz': 'o‘zbek',
        'zh_TW': 'xitoy (Tayvan)'},
 'zh_TW': {'ast': '阿斯圖里亞文',
           'bg': '保加利亞文',
           'bs': '波士尼亞文',
           'cs': '捷克文',
           'de': '德文',
           'el': '希臘文',
           'en_GB': '英文（英國）',
           'es': '西班牙文',
           'fi': '芬蘭文',
           'fr': '法文',
           'gl': '加利西亞文',
           'he': '希伯來文',
           'it': '義大利文',
           'kab': '卡比爾文',
           'ms': '馬來文',
           'pl': '波蘭文',
           'pt_BR': '葡萄牙文（巴西）',
           'ru': '俄文',
           'uk': '烏克蘭文',
           'uz': '烏茲別克文',
           'zh_TW': '中文（台灣）'}}
)




# -*- coding: utf-8 -*-

#  Copyright © 2013-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import deque
from math import atan2, sin, cos, pi, sqrt
import weakref
import itertools

import pybiklib.utils
from pybiklib.utils import epsdigits, epsilon


sid_cache = weakref.WeakKeyDictionary()

def roundeps(val):
    return round(val, epsdigits+1)
    
    
class Vector (pybiklib.utils.Vector):
    def rounded(self):
        return self.__class__(roundeps(s) for s in self)
        
    def scaled(self, vector):
        return self.__class__(s*v for s,v in zip(self, vector))
        
    def equalfuzzy(self, other):
        return all((-epsilon<s<epsilon) for s in self-other)
        
    def inversfuzzy(self, other): return (self+other).length_squared() < epsilon
        
    @classmethod
    def polygon(cls, point, n):
        x0, y0 = point
        yield x0, y0
        r = cls(point).length()
        angle0 = atan2(y0, x0)
        angled = 2 * pi / n
        
        for k in range(1, n):
            angle = angle0 + k * angled
            yield r * cos(angle), r * sin(angle)
        
        
def distance_axis_vert(axis, vert):
    ''' axis:  vector that defines a line through (0,0,0)
        return: the shortest distance between axis and vert
    '''
    axis = Vector(axis)
    base_point = axis * vert.point.dot(axis) / axis.dot(axis)
    return (vert.point - base_point).length()
    
def distance_axis_edge(axis, verts):
    ''' axis:  vector that defines a line through (0,0,0)
        verts: two points that define a line segment
        return: the shortest distance between axis and edge
    '''
    vec_axis = axis             # S1.P1 - S1.P0, S1.P0 == (0,0,0)
    vec_edge = verts[1].point - verts[0].point  # S2.P1 - S2.P0
    vec_vert0 = -verts[0].point  # S1.P0 - S2.P0
    a = vec_axis.dot(vec_axis)  # always >= 0
    b = vec_axis.dot(vec_edge)
    c = vec_edge.dot(vec_edge)  # always >= 0
    d = vec_axis.dot(vec_vert0)
    e = vec_edge.dot(vec_vert0)
    D = a*c - b*b   # always >= 0
    sD = D      # sc = sN / sD, default sD = D >= 0
    tD = D      # tc = tN / tD, default tD = D >= 0
    
    # compute the line parameters of the two closest points
    if D < epsilon: # the lines are almost parallel
        sN = 0.0    # force using point P0 on segment S1
        sD = 1.0    # to prevent possible division by 0.0 later
        tN = e
        tD = c
    else:   # get the closest points on the infinite lines
        sN = b*e - c*d
        tN = a*e - b*d
        
    if tN < 0.0:  # tc < 0 => the t=0 edge is visible
        tN = 0.0
        # recompute sc for this edge
        sN = -d
        sD = a
    elif tN > tD:   # tc > 1  => the t=1 edge is visible
        tN = tD
        # recompute sc for this edge
        sN = (-d + b)
        sD = a
    # finally do the division to get sc and tc
    sc = 0.0 if abs(sN) < epsilon else sN / sD
    tc = 0.0 if abs(tN) < epsilon else tN / tD
    
    # get the difference of the two closest points
    dP = vec_vert0 + (vec_axis * sc) - (vec_edge * tc)    # =  S1(sc) - S2(tc)
    
    return dP.length()     # return the closest distance
    
    
class Plane:
    def __init__(self, normal):
        self.normal = Vector(normal)
        self.distance = 0
        
    def __repr__(self):
        return '{0}({1.normal}, {1.distance})'.format(self.__class__.__name__, self)
        
    def signed_distance(self, point):
        return self.normal.dot(point) - self.distance
        
    def intersect_segment(self, p, q):
        pvalue = self.signed_distance(p.point)
        qvalue = self.signed_distance(q.point)
        if abs(pvalue - qvalue) < epsilon:
            return -1, None
        if abs(pvalue) < epsilon:
            return 0, p
        if abs(qvalue) < epsilon:
            return 0, q
        if pvalue < 0 < qvalue or qvalue < 0 < pvalue:
            d = pvalue / (pvalue - qvalue)
            return 1, Vert(p.point + (q.point - p.point) * d)
        else:
            return -1, None
            
        
class GeomBase:
    def __euler(self):
        chars = 'vefc456'
        halfsets, fullsets = self.get_all_ldims()
        part = lambda c,h,f: ('{}={}'.format(c,f) if h==f else '{}={}/{}'.format(c,f,h))
        return ' '.join(part(c,len(hs),len(fs)) for c,hs,fs in zip(chars,halfsets,fullsets))
        
    def __str__(self):
        return '{}({})'.format(self.__class__.__name__, self.__euler())
        
    def __repr__(self):
        sid = sid_cache.setdefault(self, len(sid_cache))
        return '{}({}, {})'.format(self.__class__.__name__, sid, self.__euler())
        
    def clone(self, clone_data):
        if id(self) in clone_data:
            return
        obj = clone_data[id(self)] = self.__class__()
        vals = [(attr, getattr(self, attr)) for attr in reversed(self.fields_geom)]
        def func(clone_data):
            for attr in self.fields_kwarg:
                if attr not in self.fields_geom:
                    setattr(obj, attr, getattr(self, attr))
            for attr, val in vals:
                if isinstance(val, list):
                    newval = [clone_data[id(o)] for o in val]
                elif val is not None:
                    newval = clone_data[id(val)]
                else:
                    newval = None
                setattr(obj, attr, newval)
        for attr, val in vals:
            if isinstance(val, list):
                for o in val:
                    yield o.clone
            elif val is not None:
                yield val.clone
        yield func
        
    def center(self):
        c = Vector()
        for i, v in enumerate(self.verts):
            c += v.point
        return c / (i+1)
        
        
class HalfGeom (GeomBase):
    fields_geom = '_ldims', '_hdim', '_fullgeom'
    
    def __init__(self, ldims, fullgeom):
        self._ldims = []
        for l in ldims:
            l.hdim = self
        #assert all(l.dim+1==self.dim for l in self._ldims)
        self._hdim = None
        self._fullgeom = None
        self.fullgeom = fullgeom
        
    def assert_valid(self):
        assert all(l.dim+1==self.dim for l in self._ldims)
        assert all(l._hdim is self for l in self._ldims)
        assert self._hdim is None or self in self._hdim._ldims
        if self.fullgeom is not None:
            assert self.fullgeom.assert_valid()
        return True
        
    @property
    def fullgeom(self):
        return self._fullgeom
    @fullgeom.setter
    def fullgeom(self, fg):
        if self._fullgeom is not None:
            self._fullgeom._halfgeoms.remove(self)
        self._fullgeom = fg
        if fg is not None:
            assert self not in fg._halfgeoms
            fg._halfgeoms.append(self)
            
    @property
    def ldims(self):
        return self._ldims
        
    @property
    def hdim(self):
        return self._hdim
    @hdim.setter
    def hdim(self, hf):
        if self._hdim is not None:
            self._hdim._ldims.remove(self)
        self._hdim = hf
        if hf is not None:
            assert self not in hf._ldims
            hf._ldims.append(self)
            
    def insert_before(self, hg):
        #TODO: try to remove this function, the order shouldnt matter,
        #      but currently needed to preserve content of model data files
        if self._hdim is not None:
            self._hdim._ldims.remove(self)
        self._hdim = hg._hdim
        assert self not in hg._hdim._ldims
        i = hg._hdim._ldims.index(hg) or len(hg._hdim._ldims)
        hg._hdim._ldims.insert(i, self)
        
    def insert_after(self, hg):
        #TODO: try to remove this function, the order shouldnt matter,
        #      but currently needed to preserve content of model data files
        if self._hdim is not None:
            self._hdim._ldims.remove(self)
        self._hdim = hg._hdim
        assert self not in hg._hdim._ldims
        i = hg._hdim._ldims.index(hg)
        hg._hdim._ldims.insert(i+1, self)
        
    def insert_after_first(self, hg):
        #TODO: try to remove this function, the order shouldnt matter,
        #      but currently needed to preserve content of model data files
        if hg._ldims:
            self.insert_after(hg._ldims[0])
        else:
            self._hdim = hg
            
    def get_all_ldims(self, halfsets=None, fullsets=None):
        if halfsets is None:
            halfsets = [set() for i in range(self.dim+1)]
        if fullsets is None:
            fullsets = [set() for i in range(self.dim+1)]
        for ldim in self._ldims:
            ldim.get_all_ldims(halfsets, fullsets)
        halfsets[self.dim].add(self)
        fullsets[self.dim].add(self._fullgeom)
        return halfsets, fullsets
        
    def rotate_ldims_to(self, hg):
        i = self._ldims.index(hg)
        if i:
            self._ldims = self._ldims[i:] + self._ldims[:i]
        return self
            
    @property
    def other(self):
        '''other halfgeom of fullgeom in the same hdim'''
        for hg in self.fullgeom.halfgeoms:
            if hg is not self and hg.hdim.hdim is self.hdim.hdim:
                return hg
        return None  # hole in surface of self.hdim.hdim
        
        
class FullGeom (GeomBase):
    fields_geom = '_halfgeoms',
    
    def __init__(self):
        self._halfgeoms = []
        
    def assert_valid(self):
        assert all(hg.fullgeom is self for hg in self.halfgeoms)
        return True
        
    @property
    def halfgeoms(self):
        return self._halfgeoms
        
    def get_all_ldims(self):
        halfsets = fullsets = None
        for hg in self._halfgeoms:
            halfsets, fullsets = hg.get_all_ldims(halfsets, fullsets)
        return halfsets, fullsets
        
        
class HalfVert (HalfGeom):
    dim = 0
    fields_arg = ()
    fields_kwarg = ()
    
    def __init__(self, *, vert=None):
        super().__init__([], vert)
        
    vert = HalfGeom.fullgeom
    halfedge = HalfGeom.hdim
    
    def assert_valid(self):
        super().assert_valid()
        return True
        
        
class Vert (FullGeom):
    fields_arg = ()
    fields_kwarg = 'point', 'id',
    
    def __init__(self, point=None, *, id=None):
        super().__init__()
        self.point = None if point is None else Vector(point)
        self.id = id
        
    def assert_valid(self):
        super().assert_valid()
        import numbers
        assert all(isinstance(i, numbers.Number) for i in self.point)
        return True
        
    halfverts = FullGeom.halfgeoms
    
        
class HalfEdge (HalfGeom):
    dim = 1
    fields_arg = ()
    fields_kwarg = ()
    
    def __init__(self, verts=None, edge=None):
        #XXX: different from other Half* classes where Half*-objects are used for initialisation
        verts = list(verts or [])
        assert all(isinstance(v, FullGeom) for v in verts)
        halfverts = [HalfVert(vert=v) for v in verts]
        super().__init__(halfverts, edge)
        
    halfverts = HalfGeom.ldims
    edge = HalfGeom.fullgeom
    halfface = HalfGeom.hdim
            
    @property
    def verts(self):
        return tuple(hv.vert for hv in self.halfverts)
        
    def create_other(self):
        return self.__class__(reversed(self.verts), self.edge)
        
    def assert_valid(self):
        assert super().assert_valid()
        assert all(isinstance(hv, HalfVert) for hv in self.halfverts)
        assert isinstance(self.edge, Edge)
        assert isinstance(self.halfface, HalfFace)
        assert all(hv.assert_valid() for hv in self.halfverts)
        assert len(self.halfverts) == 2
        assert self.edge.assert_valid()
        assert self.halfverts[1].vert is self.nextatface.halfverts[0].vert
        assert self.other is None or isinstance(self.other, self.__class__), self.other.__class__
        assert self.other is None or self.other.other is self
        return True
        
    @property
    def nextatface(self):
        ldims = self.hdim.ldims
        return ldims[(ldims.index(self)+1) % len(ldims)]
        
    @classmethod
    def polygon(cls, points, ids=None):
        vert0 = Vert(next(points))
        vertp = vert0
        ids = itertools.repeat(None) if ids is None else iter(ids)
        for point, eid in zip(points, ids):
            vertc = Vert(point)
            yield cls((vertp, vertc), Edge(id=eid))
            vertp = vertc
        yield cls((vertp, vert0), Edge(id=next(ids)))
        
    def close_hole(self, nedges, id=None):
        def next_at_hole(v):
            for hv in v.halfverts:
                he = hv.halfedge
                if he.halfface is not hf and he.other is None:
                    # only works if at the vert is only one hole
                    return he
            return None
        assert self.other is None
        hf = HalfFace(face=Face(type='face', id=id))
        hf.halfcell = self.halfface.halfcell
        he = self
        v0 = he.halfverts[1].vert
        v = he.halfverts[0].vert
        is_attach_prev = len(v0.halfverts) > 2
        is_attach_next = len(v.halfverts) > 2
        he.create_other().halfface = hf
        # attach hf to existing edges
        while v is not v0 and is_attach_next:
            he = next_at_hole(v)
            v = he.halfverts[0].vert
            is_attach_next = len(v.halfverts) > 2
            he.create_other().halfface = hf
        # attach hf to existing edges in the other direction
        halfedges = []
        while v is not v0 and is_attach_prev:
            he = next_at_hole(v0)
            halfedges.append(he)
            v0 = he.halfverts[1].vert
            is_attach_prev = len(v0.halfverts) > 2
        # if the hole needs more faces, create some edges
        for i in range(len(hf.halfedges)+len(halfedges), nedges):
            vn = Vert() if i < nedges-1 else v0
            HalfEdge([v, vn], edge=Edge()).halfface = hf
            v = vn
        for he in halfedges:
            he.create_other().halfface = hf
        return hf
        
        
class Edge (FullGeom):
    fields_arg = ()
    fields_kwarg = 'id',
    
    def __init__(self, *, id=None):
        super().__init__()
        self.id = id
        
    def assert_valid(self):
        super().assert_valid()
        assert all(isinstance(he, HalfEdge) for he in self.halfedges)
        assert len(self.halfedges) > 0
        assert all(isinstance(he, HalfEdge) for he in self.halfedges)
        assert all(he.edge is self for he in self.halfedges)
        return True
        
    halfedges = FullGeom.halfgeoms
    
    def split(self, vert):
        verts = self.halfedges[0].verts
        edge2 = self.__class__()
        for he1 in self.halfedges:
            if he1.verts[0] == verts[0]:
                he2 = HalfEdge([vert, he1.verts[1]], edge2)
                he1.halfverts[1].vert = vert
                he2.insert_after(he1)
            else:
                assert he1.verts[1] == verts[0], (he1.verts, verts)
                he2 = HalfEdge([he1.verts[0], vert], edge2)
                he1.halfverts[0].vert = vert
                he2.insert_before(he1)
                
        
class HalfFace (HalfGeom):
    dim = 2
    fields_arg = 'halfedges',
    fields_kwarg = ()
    
    def __init__(self, *, halfedges=None, face=None):
        super().__init__(halfedges or [], face)
        
    def assert_valid(self):
        assert super().assert_valid()
        assert set(self.halfedges) == set(self.ldims), (set(self.halfedges), self.ldims)
        assert isinstance(self.face, Face)
        assert self.face.assert_valid()
        assert all(he.halfface is self for he in self.ldims)
        assert all(isinstance(he, HalfEdge) for he in self.halfedges)
        assert all(he.assert_valid() for he in self.halfedges)
        return True
        
    halfedges = HalfGeom.ldims
    face = HalfGeom.fullgeom
    halfcell = HalfGeom.hdim
        
    @property
    def edges(self):
        for he in self.halfedges:
            yield he.edge
            
    @property
    def verts(self):
        for he in self.halfedges:
            yield he.verts[0]
            
    def normal(self):
        verts = self.verts
        try:
            v1 = next(verts).point
            v2 = next(verts).point
            v3 = next(verts).point
        except StopIteration:
            raise ValueError('Too few vertices')
        return (v2-v1).cross(v3-v1).normalised()
        
    def sort(self):
        try:
            he = self.halfedges[0]
        except IndexError:
            pass
        else:
            hv_last = he.halfverts[0].other
            while he.halfverts[1] is not hv_last:
                he = he.halfverts[1].other.halfedge
                he.hdim = self
        return self
        
    @classmethod
    def polygon(cls, point, n, ids=None):
        points = Vector.polygon(point, n)
        halfedges = HalfEdge.polygon(points, ids)
        halfface = cls(face=Face())
        
        for he in halfedges:
            he.halfface = halfface
        return halfface
        
    def extrude_y(self, upy, downy, ids=None):
        cell = Cell()
        self.halfcell = cell
        # down face
        face_down = Face()
        if ids is not None:
            self.face.type = face_down.type = 'face'
            self.face.id, face_down.id = ids
        halfface_down = HalfFace(face=face_down)
        
        def extrude_vert(vert_up, upy, downy):
            x1, y1 = vert_up.point
            vert_up.point = Vector((x1, upy, -y1))
            vert_down = Vert((x1, downy, -y1))
            return vert_up, vert_down
            
        vp_down = None
        for he_up in self.halfedges:
            # vertices
            vc_up, vc_down = extrude_vert(he_up.verts[1], upy, downy)
            # edges
            e_down = Edge()
            ec_side = Edge()
            if vp_down is None:
                v1_up, v1_down = vc_up, vc_down
                e0_down = e_down
                e1_side = ec_side
            else:
                # down halfedge
                HalfEdge([vc_down, vp_down], e_down).halfface = halfface_down
                # side face
                halfface_side = HalfFace(face=Face(type='face', id=he_up.edge.id))
                HalfEdge([vp_down, vc_down], e_down).halfface = halfface_side
                HalfEdge([vc_down, vc_up], ec_side).halfface = halfface_side
                HalfEdge([vc_up, vp_up], he_up.edge).halfface = halfface_side
                HalfEdge([vp_up, vp_down], ep_side).halfface = halfface_side
                halfface_side.halfcell = cell
            # next
            vp_up, vp_down = vc_up, vc_down
            ep_side = ec_side
        he_up = self.ldims[0]
        # joining down halfedge
        he_down = HalfEdge([v1_down, vp_down], e0_down)
        he_down.halfface = halfface_down
        halfface_down.rotate_ldims_to(he_down).sort()
        # joining side face
        halfface_side = HalfFace(face=Face(type='face', id=he_up.edge.id))
        HalfEdge([vp_down, v1_down], e0_down).halfface = halfface_side
        HalfEdge([v1_down, v1_up], e1_side).halfface = halfface_side
        HalfEdge([v1_up, vp_up], he_up.edge).halfface = halfface_side
        HalfEdge([vp_up, vp_down], ep_side).halfface = halfface_side
        halfface_side.halfcell = cell
        
        halfface_down.halfcell = cell
        return cell
        
    def pyramid(self, upy, downy, id=None):
        cell = Cell()
        self.halfcell = cell
        vert_up = Vert((0., upy, 0.))
        
        def change_vert(vert_down, downy):
            x1, y1 = vert_down.point
            vert_down.point = Vector((-x1, downy, -y1))
            return vert_down
            
        vp_down = None
        for he in self.halfedges:
            # vertices
            vc_down = change_vert(he.verts[1], downy)
            # edges
            ec_side = Edge()
            if vp_down is None:
                v1_down = vc_down
                e1_side = ec_side
            else:
                # side face
                halfface_side = HalfFace(face=Face(type='face', id=he.edge.id))
                HalfEdge([vc_down, vp_down], he.edge).halfface = halfface_side
                HalfEdge([vp_down, vert_up], ep_side).halfface = halfface_side
                HalfEdge([vert_up, vc_down], ec_side).halfface = halfface_side
                halfface_side.halfcell = cell
            # next
            vp_down = vc_down
            ep_side = ec_side
        he = self.ldims[0]
        # joining side face
        halfface_side = HalfFace(face=Face(type='face', id=he.edge.id))
        HalfEdge([v1_down, vp_down], he.edge).halfface = halfface_side
        HalfEdge([vp_down, vert_up], ep_side).halfface = halfface_side
        HalfEdge([vert_up, v1_down], e1_side).halfface = halfface_side
        halfface_side.halfcell = cell
        # down face
        if id is not None:
            self.face.type = 'face'
            self.face.id = id
        return cell
        
        
class Face (FullGeom):
    fields_arg = ()
    fields_kwarg = 'type', 'id'
    
    def __init__(self, *, type=None, id=None):
        super().__init__()
        self.type = type
        self.id = id
        
    def assert_valid(self):
        super().assert_valid()
        assert all(isinstance(f, HalfFace) for f in self.halffaces)
        assert all(hf.face is self for hf in self.halffaces)
        assert 1 <= len(self.halffaces) <= 2
        lenhf0 = len(list(self.halffaces[0].halfedges))
        assert all(lenhf0 == len(list(hf.halfedges)) for hf in self.halffaces)
        return True
        
    halffaces = FullGeom.halfgeoms
    
    @property
    def edges(self):
        return self.halffaces[0].edges
        
    @property
    def verts(self):
        return self.halffaces[0].verts
        
    def split(self, split_verts):
        assert len(split_verts) == 2, len(split_verts)
        edge = Edge()
        newface = Face(type=self.type, id=self.id)
        edgef1 = None
        for halfface in self.halffaces:
            he = halfface.ldims[0]
            # make sure to start on the same edge for all halffaces
            if edgef1 is None:
                edgef1 = he.edge
            else:
                he = next(_he for _he in edgef1.halfedges if _he.halfface is halfface)
                halfface.rotate_ldims_to(he)
            hedge2f1 = None
            hedge1f2 = None
            for he in halfface.halfedges[:]:
                if hedge2f1 is None:
                    if he.verts[1] in split_verts:
                        # end of the old halfface
                        hedge2f1 = he
                elif hedge1f2 is None:
                    # start of the new halfface
                    newhalfface = HalfFace(face=newface)
                    newhalfface.insert_after(halfface)
                    he.halfface = newhalfface
                    hedge1f2 = he
                elif he.verts[1] not in split_verts:
                    # proceed to the end of the new halfface
                    he.halfface = newhalfface
                else:
                    # end of the new halfface
                    he.halfface = newhalfface
                    newhalfface.rotate_ldims_to(hedge1f2)
                    hedge2f2 = he
                    break
            # insert a new halfedge between the splitted halfface
            hedge1 = HalfEdge(split_verts, edge)
            hedge2 = HalfEdge([split_verts[1], split_verts[0]], edge)
            if hedge2f1.verts[1] is split_verts[1]:
                hedge1, hedge2 = hedge2, hedge1
            hedge1.insert_after(hedge2f1)
            hedge2.insert_after(hedge2f2)
        return edge
        
        
class HalfCell (HalfGeom):
    dim = 3
    fields_arg = 'halffaces',
    fields_kwarg = 'indices',
    
    def __init__(self, *, halffaces=None):
        super().__init__(halffaces or [], None)
        self.indices = None
        
    def assert_valid(self):
        assert super().assert_valid()
        assert list(self.halffaces) == list(self.ldims), (list(self.halffaces), list(self.ldims))
        assert all(isinstance(f, HalfFace) for f in self.halffaces)
        assert all(hf.assert_valid() for hf in self.halffaces)
        assert len(list(self.edges)) > 3
        assert len(list(self.verts)) > 3
        assert all(he.hdim.hdim is not None for e in self.edges for he in e.halfedges)
        return True
        
    halffaces = HalfGeom.ldims
                
    @property
    def faces(self):
        for hf in self.halffaces:
            yield hf.face
            
    @property
    def edges(self):
        cache = []
        for f in self.faces:
            for e in f.edges:
                if e not in cache:
                    cache.append(e)
                    yield e
                    
    @property
    def verts(self):
        cache = []
        for hf in self.halffaces:
            for v in hf.verts:
                if v not in cache:
                    cache.append(v)
                    yield v
                    
    def split(self, split_edges, id=None):
        halffaces, halffaces1, halffaces2 = [self.ldims[0]], [self.ldims[0]], []
        # new face
        face = Face(type='cut', id=id)
        halfface1 = HalfFace(face=face)
        halfface2 = HalfFace(face=face)
        
        while halffaces:
            for he in halffaces.pop().halfedges:
                hf = he.other.halfface
                if he.edge in split_edges:
                    he.create_other().halfface = halfface1
                    if hf not in halffaces2:
                        halffaces2.append(hf)
                else:
                    if hf not in halffaces1:
                        halffaces.append(hf)
                        halffaces1.append(hf)
        halffaces = halffaces2[:]
        while halffaces:
            for he in halffaces.pop().halfedges:
                if he.edge in split_edges:
                    he.create_other().halfface = halfface2
                else:
                    hf = he.other.halfface
                    if hf not in halffaces2:
                        halffaces.append(hf)
                        halffaces2.append(hf)
        # update cell
        halfface1.sort()
        phf = halfface1
        for hf in halffaces1:
            phf = hf
            #XXX: just to reorder ldims to the order of halfface1, try to remove this line:
            hf.hdim = self
        halfface1.hdim = self
        
        # new cell
        halfface2.sort()
        halffaces2.append(halfface2)
        return Cell(halffaces=halffaces2)
            
    @classmethod
    def dodecahedron(cls, ids=None):
        #rfi = sqrt(25 + 10*sqrt(5)) / 10
        rfu = sqrt(50 + 10*sqrt(5)) / 10
        ri = sqrt((25 + 11*sqrt(5))/10) / 2
        #ru = sqrt(3) * (1 + sqrt(5)) / 4
        
        #1: h01² + (r1u-rfu)² = 1²
        #2: (ri-h01)² + r1u² = ru²
        # h01 = rfu
        # r1u² = sqrt(45 + 20*sqrt(5)) / 5
        r1u = sqrt(sqrt(45 + 20*sqrt(5)) / 5)
        
        ids = itertools.repeat(None) if ids is None else iter(ids)
        cell = cls()
        
        hf_down = HalfFace.polygon((0.,-rfu), 5)
        hf_down.halfcell = cell
        hf_down.face.type = 'face'
        hf_down.face.id = next(ids)
        for v in hf_down.verts:
            x,z = v.point
            v.point = Vector((x, -ri, z))
        hfs1 = [he.close_hole(5, id=fid) for he, fid in zip(hf_down.halfedges, ids)]
        hfs2 = [hf.halfedges[2].close_hole(5, id=fid) for hf, fid in zip(hfs1, ids)]
        hf_up = hfs2[0].halfedges[3].close_hole(5, id=next(ids))
        
        hes01 = [he for e in cell.edges for he in e.halfedges if he.verts[0].point is not None and he.verts[1].point is None]
        for he in hes01:
            v0 = he.verts[0]
            v1 = he.verts[1]
            x,y,z = v0.point
            v1.point = Vector((x/rfu*r1u, rfu-ri, z/rfu*r1u))
        hes12 = [he for e in cell.edges for he in e.halfedges if he.verts[0].point is not None and he.verts[1].point is None]
        for he in hes12:
            he.verts[1].point = True  #XXX: marker, any value not None would do it
        hes23 = [he for e in cell.edges for he in e.halfedges if he.verts[0].point is not None and he.verts[1].point is None]
        for i, he in enumerate(hes23):
            he.verts[0].point = -hes01[(i+2)%5].verts[1].point
            he.verts[1].point = -hes01[(i+2)%5].verts[0].point
            
        return cell
        
Cell = HalfCell


class Polyhedron:
    def __init__(self):
        self._cells = []
        
    def clone(self):
        clone_data = {}
        obj = self.__class__()
        funcs = deque(f for o in self._cells for f in o.clone(clone_data))
        obj._cells = [clone_data[id(o)] for o in self._cells]
        while funcs:
            f = funcs.popleft()
            r = f(clone_data)
            if r is not None:
                funcs.extendleft(reversed(list(r)))
        return obj
        
    def assert_valid(self):
        import itertools
        assert all(isinstance(c, Cell) and c.assert_valid() for c in self.cells)
        assert all(isinstance(f, Face) and f.assert_valid() for f in self.faces)
        assert all(isinstance(e, Edge) and e.assert_valid() for e in self.edges)
        assert all(isinstance(v, Vert) and v.assert_valid() for v in self.verts)
        assert all(not v1.point.equalfuzzy(v2.point) for v1, v2 in itertools.combinations(self.verts, 2)), self.verts
        return True
        
    @property
    def cells(self):
        return self._cells
    @cells.setter
    def cells(self, cells):
        self._cells = cells
        
    @property
    def faces(self):
        cache = []
        for c in self.cells:
            for f in c.faces:
                if f not in cache:
                    cache.append(f)
        return cache
        
    @property
    def edges(self):
        cache = []
        for f in self.faces:
            for e in f.edges:
                if e not in cache:
                    cache.append(e)
        return cache
        
    @property
    def verts(self):
        cache = []
        for f in self.faces:
            for v in f.verts:
                if v not in cache:
                    cache.append(v)
        return cache
        
    def set_verts(self, *args):
        self._polytopes = [Vert(arg) for arg in args]
            
    def set_edges(self, *args):
        halfedges = []
        for arg in args:
            verts = [self._polytopes[i] for i in arg]
            halfedges.append(HalfEdge(verts=verts, edge=Edge()))
        self._polytopes = halfedges
        
    def set_faces(self, *args, ids=None):
        halffaces = []
        if ids is None:
            faces = [Face() for unused in args]
        else:
            assert len(args) == len(ids)
            faces = [Face(type='face', id=fid) for fid in ids]
        for arg, face in zip(args, faces):
            halfface = HalfFace(face=face)
            for i in arg:
                # this only works if args describe one closed cell
                if i > 0:
                    self._polytopes[i-1].halfface = halfface
                elif i < 0:
                    self._polytopes[-i-1].create_other().halfface = halfface
                else:
                    raise ValueError()
            halffaces.append(halfface)
        del self._polytopes
        self._cells = [Cell(halffaces=halffaces)]
            
    def split_plane(self, plane, indexpos=None, split_id=None):
        # split intersected edges
        new_verts = []
        for edge in self.edges[:]:
            pos, vert = plane.intersect_segment(*edge.halfedges[0].verts)
            if pos > 0:
                edge.split(vert)
                new_verts.append(vert)
            elif pos == 0 and vert not in new_verts:
                new_verts.append(vert)
        # split intersected faces
        new_edges = []
        for face in self.faces[:]:
            split_verts = [v for v in face.verts for nv in new_verts if nv is v]
            assert 0 <= len(split_verts) <= 2, split_verts
            if len(split_verts) == 2:
                for he in face.halffaces[0].halfedges:
                    if set(he.verts) == set(split_verts):
                        new_edges.append(he.edge)
                        break
                else:
                    new_edge = face.split(split_verts)
                    new_edges.append(new_edge)
        # split cells
        for cell in self.cells[:]:
            split_edges = [e for e in cell.edges if e in new_edges]
            if len(split_edges) > 2:
                new_cell = cell.split(split_edges, split_id)
                self.cells.append(new_cell)
            else:
                new_cell = None
            if indexpos is not None:
                for hf in cell.halffaces:
                    for he in hf.halfedges:
                        if he.edge not in split_edges:
                            sd1 = -plane.signed_distance(he.verts[0].point)
                            # the verts of the edge e cannot be both in range epsilon to the plane,
                            # otherwise e would be in split_edges, so only test one vert for epsilon
                            if sd1 > epsilon or sd1 >= -epsilon and plane.signed_distance(he.verts[1].point) < 0:
                                if new_cell is not None:
                                    new_cell.indices = cell.indices
                                new_cell = cell
                            elif new_cell is None:
                                break
                            indices = list(cell.indices)
                            indices[indexpos] += 1
                            new_cell.indices = tuple(indices)
                            break
                    else:
                        continue
                    break
            
    def distances(self, plane):
        ddict = {}
        for v in self.verts:
            dist = plane.signed_distance(v.point)
            ddict.setdefault(roundeps(dist), []).append(dist)
        return sorted(sum(dists)/len(dists) for dists in ddict.values())
        
    def scale(self, value):
        for v in self.verts:
            v.point = v.point.scaled(value)
            
    def remove_cells(self, func):
        self._cells = [c for c in self._cells if not func(c)]
        
    

